r"""Shallow water 2 layers solver (:mod:`fluidsim_ocean.sw2l.solver`)
====================================================================

.. autoclass:: InfoSolverSW2L
   :members:
   :private-members:

.. autoclass:: Simul
   :members:
   :private-members:

"""

from fluidsim.base.setofvariables import SetOfVariables
from fluidsim.base.solvers.pseudo_spect import (
    SimulBasePseudoSpectral,
    InfoSolverPseudoSpectral,
)

from fluiddyn.util import mpi

class InfoSolverSW2L(InfoSolverPseudoSpectral):
    """Information about the solver SW1L."""

    def _init_root(self):
        """The simulation object is instantiated with classes defined in this
        function.

        The super function `InfoSolverPseudoSpectral._init_root` is
        called first. A few classes defined by this function are retained as it
        is:

        - :class:`fluidsim.base.time_stepping.pseudo_spect_cy.TimeSteppingPseudoSpectral`

        - :class:`fluidsim.solvers.sw1l.OperatorsPseudoSpectralSW1L`

        - :class:`fluidsim.base.preprocess.pseudo_spect.PreprocessPseudoSpectral`

        Solver-specific first-level classes are added:

        - :class:`fluidsim.solvers.sw1l.solver.Simul`

        - :class:`fluidsim.solvers.sw1l.state.StateSW1L`

        - :class:`fluidsim.solvers.sw1l.init_fields.InitFieldsSW1L`

        - :class:`fluidsim.solvers.sw1l.output.OutputSW1L`

        - :class:`fluidsim.solvers.sw1l.forcing.ForcingSW1L`

        """
        super(InfoSolverSW1L, self)._init_root()

        package = "fluidsim_ocean.sw2l"

        self.module_name = package + ".solver"
        self.class_name = "Simul"
        self.short_name = "sw2l"

        classes = self.classes

        # classes.Operators.module_name = package + ".operators"
        # classes.Operators.class_name = "OperatorsPseudoSpectralSW1L"

        classes.State.module_name = package + ".state"
        classes.State.class_name = "StateSW1L"

        # classes.InitFields.module_name = package + ".init_fields"
        # classes.InitFields.class_name = "InitFieldsSW2L"

        # classes.Output.module_name = package + ".output"
        # classes.Output.class_name = "OutputSW2L"

        # classes.Forcing.module_name = package + ".forcing"
        # classes.Forcing.class_name = "ForcingSW2L"


class Simul(SimulBasePseudoSpectral):
    """A solver of the shallow-water 1 layer equations (SW1L).

    .. inheritance-diagram:: Simul

    """

    InfoSolver = InfoSolverSW2L

    @staticmethod
    def _complete_params_with_default(params):
        r"""This static method is used to complete the *params* container.

        Notes
        -----

        Which input parameters do we want/need?

        - :math:`f` is the system rotation.

        - ???

        """
        SimulBasePseudoSpectral._complete_params_with_default(params)

        attribs = {"f": 0}
        params._set_attribs(attribs)

    def __init__(self, params):
        # Parameter(s) specific to this solver

        super(Simul, self).__init__(params)

        # if mpi.rank == 0:
        #     self.output.print_stdout(
        #         "c2 = {0:6.5g} ; f = {1:6.5g} ; kd2 = {2:6.5g}".format(
        #             params.c2, params.f, params.kd2
        #         )
        #     )

    def tendencies_nonlin(self, state_spect=None, old=None):
        r"""Compute the nonlinear tendencies.

        Parameters
        ----------

        state_spect : :class:`fluidsim.base.setofvariables.SetOfVariables`
            optional

            Array containing the state, i.e. the horizontal velocities and
            surface displacement scalar in Fourier space.  When `state_spect` is
            provided, the variables vorticity and the velocities and surface
            displacement are computed from it, otherwise, they are taken from
            the global state of the simulation, `self.state`.

            These two possibilities are used during the Runge-Kutta
            time-stepping.

        Returns
        -------

        tendencies_fft : :class:`fluidsim.base.setofvariables.SetOfVariables`
            An array containing the tendencies for the velocities
            :math:`\mathbf u` and the surface displacement scalar :math:`\eta`.

        Notes
        -----

        .. |uu| mathmacro:: \mathbf{u}
        .. |p| mathmacro:: \partial

        The two-layer shallow water equations can be written as:

        ... First, we need to write here the simplest form of the governing
        equations (first, without dissipating terms). With all variables defined
        :-).

        .. math::

           \p_t \uu_0(x,y) &= ...

           \p_t \uu_1(x,y) &= ...

        .. math::

           \p_t h_n(x,y) + \nabla \cdot (\mathbf{u}_n h_n ) = 0.

        This function computes all terms on the RHS of the equations above,
        including the nonlinear term. The linear dissipation term (not shown
        above) is computed implicitly, as demonstrated in
        :class:`fluidsim.base.time_stepping.pseudo_spect.TimeSteppingPseudoSpectral`.

        Main conserved quantities?

        We need to write the formula for the main quantities conserved by the
        equations.

        - energy: ...

        - Ertel potential vorticity: ...

        """
        oper = self.oper

        if state_spect is None:
            state_phys = self.state.state_phys
        else:
            state_phys = self.state.return_statephys_from_statespect(state_spect)

        ux = state_phys.get_var("ux")
        uy = state_phys.get_var("uy")
        eta = state_phys.get_var("eta")
        rot = state_phys.get_var("rot")

        if old is None:
            tendencies_fft = SetOfVariables(
                like=self.state.state_spect, info="tendencies_nonlin"
            )
        else:
            tendencies_fft = old

        Fx_fft = tendencies_fft.get_var("ux_fft")
        Fy_fft = tendencies_fft.get_var("uy_fft")
        Feta_fft = tendencies_fft.get_var("eta_fft")

        fft2 = oper.fft2

        F1x, F1y = oper.compute_Frot(rot, ux, uy, f)
        gradx_fft, grady_fft = oper.gradfft_from_fft(
            fft2(self.params.c2 * eta + (ux ** 2 + uy ** 2) / 2.)
        )
        oper.dealiasing(gradx_fft, grady_fft)
        Fx_fft[:] = fft2(F1x) - gradx_fft
        Fy_fft[:] = fft2(F1y) - grady_fft

        Feta_fft[:] = -oper.divfft_from_vecfft(
            fft2((eta + 1) * ux),
            fft2((eta + 1) * uy)
        )

        oper.dealiasing(tendencies_fft)

        if self.params.forcing.enable:
            tendencies_fft += self.forcing.get_forcing()

        return tendencies_fft


if __name__ == "__main__":

    import numpy as np

    import fluiddyn as fld

    params = Simul.create_default_params()

    params.short_name_type_run = "test"

    nh = 32
    Lh = 2 * np.pi
    params.oper.nx = nh
    params.oper.ny = nh
    params.oper.Lx = Lh
    params.oper.Ly = Lh

    delta_x = params.oper.Lx / params.oper.nx
    params.nu_8 = (
        2. * 10e-1 * params.forcing.forcing_rate ** (1. / 3) * delta_x ** 8
    )

    params.time_stepping.t_end = 1.
    # params.time_stepping.USE_CFL = False
    # params.time_stepping.deltat0 = 0.01

    params.init_fields.type = "noise"

    params.forcing.enable = True
    params.forcing.type = "waves"

    params.output.periods_print.print_stdout = 0.25

    params.output.periods_save.phys_fields = 0.5
    params.output.periods_save.spectra = 0.5
    params.output.periods_save.spect_energy_budg = 0.5
    params.output.periods_save.increments = 0.5
    params.output.periods_save.pdf = 0.5
    params.output.periods_save.time_signals_fft = True

    params.output.periods_plot.phys_fields = 0.

    params.output.phys_fields.field_to_plot = "eta"

    sim = Simul(params)

    sim.output.phys_fields.plot(key_field="eta")
    sim.time_stepping.start()
    sim.output.phys_fields.plot(key_field="eta")

    fld.show()
